/*
 * Copyright (C) 2010 - Paulo Ricardo Zanoni <pzanoni@mandriva.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 */

#include <iostream>

#include "ExtendDesktopWidget.h"

ExtendDesktopWidget::ExtendDesktopWidget(QWidget *parent)
    : QWidget(parent)
{
    QVBoxLayout *v_lay = new QVBoxLayout(this);
    QHBoxLayout *h_lay = new QHBoxLayout();
    QLabel *label = new QLabel(tr("Output layout:"));

    left_btn = new QPushButton("<<");
    right_btn = new QPushButton(">>");

    table = new QTableWidget(1,1);
    extend_box = new QCheckBox(tr("Extended desktop"));

    table->verticalHeader()->hide();
    table->horizontalHeader()->hide();
    table->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    table->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    table->setSelectionMode(QAbstractItemView::SingleSelection);

    QString string = "--";
    add_table_item(0, string, None);
    update_table_size();

    left_btn->setFixedWidth(40);
    right_btn->setFixedWidth(40);

    h_lay->setAlignment(Qt::AlignLeft);
    h_lay->addWidget(left_btn);
    h_lay->addWidget(right_btn);

    v_lay->addWidget(extend_box);
    v_lay->addWidget(label);
    v_lay->addWidget(table);
    v_lay->addLayout(h_lay);

    connect(left_btn, SIGNAL(clicked()), this, SLOT(left_btn_click()));
    connect(right_btn, SIGNAL(clicked()), this, SLOT(right_btn_click()));

}

ExtendDesktopWidget::~ExtendDesktopWidget()
{
    // XXX: fix this
}

void ExtendDesktopWidget::update_table_size()
{
    int frame_width = table->frameWidth();
    table->resizeColumnsToContents();
    table->resizeRowsToContents();
    table->setFixedSize(table->horizontalHeader()->length() + (2 * frame_width),
			table->verticalHeader()->length() + (2 * frame_width));

    // XXX: yes, I know this could be somewhere else:
    if (table->columnCount() < 2) {
        extend_box->setEnabled(false);
        left_btn->setEnabled(false);
        right_btn->setEnabled(false);
        table->setEnabled(false);
    } else {
        extend_box->setEnabled(true);
        left_btn->setEnabled(true);
        right_btn->setEnabled(true);
        table->setEnabled(true);
    }
}

void ExtendDesktopWidget::move_selected_column(Direction dir)
{
    int row, col;
    int offset;
    QTableWidgetItem *curr, *other;

    if (dir == Left)
        offset = -1;
    else
        offset = +1;

    curr = table->currentItem();
    row = curr->row();
    col = curr->column();

    if (col + offset >= table->columnCount() || col + offset < 0) {
        return;
    }

    curr = table->takeItem(row, col);
    other = table->takeItem(row, col+offset);

    table->setItem(row, col+offset, curr);
    table->setItem(row, col, other);

    table->setCurrentCell(row, col+offset);

    update_table_size();
}

void ExtendDesktopWidget::left_btn_click()
{
    move_selected_column(Left);
}

void ExtendDesktopWidget::right_btn_click()
{
    move_selected_column(Right);
}

bool ExtendDesktopWidget::is_extended()
{
    if (extend_box->checkState() == Qt::Checked)
	return true;
    else
	return false;

}

void ExtendDesktopWidget::set_extended(bool extended)
{
//    if (extended)
//	std::cout << "extending and " << table->columnCount() << "\n";
//    else
//	std::cout << "not extending and " << table->columnCount() << "\n";

    if (table->columnCount() < 2) {
	extend_box->setEnabled(false);
	left_btn->setEnabled(false);
	right_btn->setEnabled(false);
	table->setEnabled(false);
    } else {
	extend_box->setEnabled(true);
	left_btn->setEnabled(true);
	right_btn->setEnabled(true);
	table->setEnabled(true);
    }

    if (extended == true) {
	extend_box->setCheckState(Qt::Checked);
    } else {
	extend_box->setCheckState(Qt::Unchecked);
    }
}

void ExtendDesktopWidget::set_items(QList<QPair<RROutput, QString*>* > items)
{
    int i;
    QPair<RROutput, QString*> *pair;

    for(i = 0; i < table->columnCount(); i++) {
	//std::cout << "Removing " << i << "\n";
	table->removeCellWidget(0, i);
    }

    if (items.size() == 0) {
	table->setColumnCount(1);
	table->item(0,0)->setText("--");
	table->item(0,0)->setData(Qt::UserRole, (uint)None);

    } else {
	table->setColumnCount(items.size());
	for(i = 0; i < items.size(); i++) {
	    pair = items.at(i);
	    //std::cout << "Adding " << pair->second->toStdString() << "\n";

	    add_table_item(i, *(pair->second), pair->first);
	}
    }

    if (items.size() >= 2) {
	//set_extended(true);
	extend_box->setEnabled(true);
	left_btn->setEnabled(true);
	right_btn->setEnabled(true);
	table->setEnabled(true);
    } else {
	set_extended(false);
    }

    update_table_size();
}

void ExtendDesktopWidget::add_output(RROutput output, std::string output_name)
{
    int i, new_column;
    QString string = output_name.c_str();

    if (table->columnCount() == 1 &&
	table->item(0,0)->data(Qt::UserRole).toUInt() == None) {
	table->item(0,0)->setText(string);
	table->item(0,0)->setData(Qt::UserRole, (uint)output);
	update_table_size();
    }

    for(i = 0; i < table->columnCount(); i++) {
	if (table->item(0, i)->data(Qt::UserRole).toUInt() == output) {
	    return;
	}
    }

    //std::cout << "adding " << output_name << "\n";
    new_column = table->columnCount();
    table->setColumnCount(new_column+1);


    add_table_item(new_column, string, output);
}

void ExtendDesktopWidget::remove_output(RROutput output)
{
    int i, j;
    int columns;
    bool removed = false;

    QTableWidgetItem *item;

    columns = table->columnCount();

    for(i = 0; i < columns; i++) {
	if (table->item(0, i)->data(Qt::UserRole).toUInt() == output) {
	    table->removeCellWidget(0, i);
	    removed = true;
	    for(j = i+1; j < columns; j++) {
		item = table->takeItem(0, j);
		table->setItem(0, j-1, item);
	    }

	    break;
	}
    }

    if (removed && columns == 1) {
	table->item(0,0)->setText("--");
	table->item(0,0)->setData(Qt::UserRole, (uint)None);
    } else if (removed) {
	table->setColumnCount(columns-1);
    }
    update_table_size();
}

QList<RROutput> ExtendDesktopWidget::get_outputs_in_order()
{
    int i;
    QList<RROutput> ret;
    RROutput output;

    for(i = 0; i < table->columnCount(); i++) {
	output = table->item(0, i)->data(Qt::UserRole).toUInt();
	if (output == None) {
	    break;
	}
	ret << output;
    }
    return ret;
}

void ExtendDesktopWidget::add_table_item(int column, QString name, RROutput id)
{
    QTableWidgetItem *item;

    item = new QTableWidgetItem(name, QTableWidgetItem::Type);
    item->setFlags(item->flags() & ~Qt::ItemIsEditable);
    item->setData(Qt::UserRole, (uint)id);
    table->setItem(0, column, item);

    update_table_size();
}
