/*
 * Copyright (C) 2010 - Paulo Ricardo Zanoni <pzanoni@mandriva.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 */

#include <iostream>

#include <QtGui>

#include "OutputWidget.h"

OutputWidget::OutputWidget(RandrInfo *randr_info_arg,
			   RROutput output_id_arg,
			   QWidget *parent)
    : QWidget(parent)
{
    QHBoxLayout *h_lay;
    QVBoxLayout *v_lay;
    QLabel *name_label;

    randr_info = randr_info_arg;
    output_id = output_id_arg;

    h_lay = new QHBoxLayout(this);
    v_lay = new QVBoxLayout();

    name_label = new QLabel(randr_info->get_output_name(output_id).c_str());

    mode_combo = new QComboBox();
    rotation_combo = new QComboBox();

    name_label->setMinimumWidth(50);
    mode_combo->setMinimumWidth(200);
    rotation_combo->setMinimumWidth(200);

    rotation_combo->addItem(tr("No rotation"), RR_Rotate_0);
    rotation_combo->addItem(tr("Rotate to left"), RR_Rotate_90);
    rotation_combo->addItem(tr("Rotate to right"), RR_Rotate_270);
    rotation_combo->addItem(tr("Rotate upside-down"), RR_Rotate_180);

    update_info();

    connect(mode_combo, SIGNAL(currentIndexChanged(int)),
	    this, SLOT(mode_combo_changed(int)));
    connect(rotation_combo, SIGNAL(currentIndexChanged(int)),
	    this, SLOT(rotation_combo_changed(int)));

    h_lay->addWidget(name_label);
    v_lay->addWidget(mode_combo);
    v_lay->addWidget(rotation_combo);
    h_lay->addLayout(v_lay);

}

OutputWidget::~OutputWidget()
{
    delete mode_combo;
    delete rotation_combo;
}

void OutputWidget::update_info()
{
    int index;
    Rotation curr_rotation;

    dont_emit_changes = true;

    mode_combo->setCurrentIndex(0);
    mode_combo->clear();

    if (! randr_info->output_is_connected(output_id)) {
	mode_combo->setEnabled(false);
	mode_combo->addItem(tr("disconnected"), (uint)None);
	selected_mode = None;

	rotation_combo->setEnabled(false);
	index = rotation_combo->findData((uint)RR_Rotate_0);
	rotation_combo->setCurrentIndex(index);
	selected_rotation = RR_Rotate_0;

    } else {
	mode_combo->setEnabled(true);
	rotation_combo->setEnabled(true);

	mode_combo->addItem(tr("off"), (uint)None);

	if (randr_info->get_output_mode(output_id) == None) {
	    mode_combo->setCurrentIndex(mode_combo->count() -1);
	    selected_mode = None;
	}

	foreach(RRMode mode, randr_info->get_output_modes(output_id)) {
	    std::string mode_name = randr_info->get_mode_name(mode);
	    //std::cout << "Mode id: " << mode << " name:" << mode_name << "\n";
	    //std::cout << "Adding item " << mode_combo->count() << "\n";
	    mode_combo->addItem(mode_name.c_str(), (uint)mode);

	    if (randr_info->get_output_mode(output_id) == mode) {
		mode_combo->setCurrentIndex(mode_combo->count() -1);
		selected_mode = mode;
	    }
	}

	curr_rotation = randr_info->get_output_rotation(output_id);
	index = rotation_combo->findData((uint)curr_rotation);
	rotation_combo->setCurrentIndex(index);
	selected_rotation = curr_rotation;
    }

    dont_emit_changes = false;
    emit output_changed(output_id, selected_mode, selected_rotation);
}

void OutputWidget::mode_combo_changed(int index)
{
    //std::cout << "New index: " << index;
    selected_mode = mode_combo->itemData(index).toUInt();

    if (!dont_emit_changes) {
	emit output_changed(output_id, selected_mode, selected_rotation);
    }
}

void OutputWidget::rotation_combo_changed(int index)
{
    selected_rotation = rotation_combo->itemData(index).toUInt();
    if (!dont_emit_changes) {
	emit output_changed(output_id, selected_mode, selected_rotation);
    }
}

RRMode OutputWidget::get_selected_mode()
{
    return selected_mode;
}

Rotation OutputWidget::get_selected_rotation()
{
    //std::cout << "Selected rotation: " << selected_rotation << "\n";
    return selected_rotation;
}

RROutput OutputWidget::get_output()
{
    return output_id;
}

void OutputWidget::set_selected_mode(RRMode mode_id)
{
    int index;

    if (!randr_info->output_is_connected(output_id)) {
	return;
    }

    selected_mode = mode_id;

    index = mode_combo->findData((uint)mode_id);
    if (index != -1) {
	mode_combo->setCurrentIndex(index);
	emit output_changed(output_id, selected_mode, selected_rotation);
    } else {
	std::cout << "OutputWidget: attempting to set an invalid mode!\n";
    }
}

void OutputWidget::set_selected_rotation(Rotation rotation)
{
    int index;

    if (!randr_info->output_is_connected(output_id)) {
	return;
    }

    selected_rotation = rotation;

    index = rotation_combo->findData((uint)rotation);
    if (index != -1) {
	rotation_combo->setCurrentIndex(index);
	emit output_changed(output_id, selected_mode, selected_rotation);
    } else {
	// This should never happen. Better safe than sorry.
	std::cout << "OutputWidget: attempting to set invalid rotation!\n";
    }
}
