/*
 * Copyright (C) 2010 - Paulo Ricardo Zanoni <pzanoni@mandriva.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 */

#include <cstdlib>

// XXX: find out why we need to include this before Xlib.h and not after!
#include <boost/foreach.hpp>

#include <X11/Xlib.h>
#include <X11/extensions/Xrandr.h>

#include "RandrInfo.h"

#include <iostream>
using namespace std;

#define MODE_NAME(id)   modes[(id)]->name
#define OUTPUT_NAME(id) outputs[(id)]->name
#define OUTPUT_CRTC(id) outputs[(id)]->crtc
#define OUTPUT_MODE(id) crtcs[OUTPUT_CRTC((id))]->mode
#define SIDE_ROTATED(rot) (rot == RR_Rotate_90 || rot == RR_Rotate_270)

typedef map<RRCrtc, RROutput> CrtcToOutputMap;

// XXX: Xlib really makes a mess with signed/unsigned integers when referring to
// window sizes. width/height in XRRCrtcInfo are unsigned, but width/height used
// in XRRSetScreenSize are signed.

RandrInfo::RandrInfo(Display *display_arg, int screen_nbr_arg)
{
    int xrr_major_version = 1;
    int xrr_minor_version = 3;

    display = display_arg;
    screen_nbr = screen_nbr_arg;

    if (!XRRQueryExtension(display, &xrr_event_base, &xrr_error_base)) {
	cerr << "Unable to query randr extension\n";
	exit(1);
    }

    if (!XRRQueryVersion(display, &xrr_major_version, &xrr_minor_version)) {
	cerr << "XRRQueryVersion returned False\n";
	exit(1);
    }

    cout << "Server major: " << xrr_major_version
	 << " minor: " << xrr_minor_version << "\n";

    res = NULL;
    update_info();
}

RandrInfo::~RandrInfo()
{
    XRRFreeScreenResources(res);

    BOOST_FOREACH(OutputMap::value_type o_it, outputs) {
	delete o_it.second;
    }

    BOOST_FOREACH(ModeMap::value_type m_it, modes) {
	delete m_it.second;
    }

    BOOST_FOREACH(CrtcMap::value_type c_it, crtcs) {
	delete c_it.second;
    }
}

void RandrInfo::update_info()
{
    int i;

    Window root_win;
    RROutput output_id;
    RRMode mode_id;
    RRCrtc crtc_id;

    cout << "-- updating RandrInfo info --\n";

    root_win = RootWindow(display, screen_nbr);

    if (res) {
	XRRFreeScreenResources(res);
    }

    res = XRRGetScreenResources(display, root_win);

    // Outputs
    for (i = 0; i < res->noutput; i++) {
	output_id = res->outputs[i];
	if (outputs.find(output_id) == outputs.end()) {
	    cout << "Creating output " << output_id << "\n";
	    outputs[output_id] = new OutputInfo(display, res, output_id);
	} else {
	    outputs[output_id]->update_info(display, res);
	}
    }


    // Modes
    for (i = 0; i < res->nmode; i++) {
	mode_id = res->modes[i].id;
	if (modes.find(mode_id) == modes.end()) {
	    cout << "Creating mode " << mode_id << "\n";
	    modes[mode_id] = new ModeInfo(&(res->modes[i]));
	} else {
	    modes[mode_id]->update_info(&(res->modes[i]));
	}
    }

    // CRTCs
    for (i = 0; i < res->ncrtc; i++) {
	crtc_id = res->crtcs[i];
	if (crtcs.find(crtc_id) == crtcs.end()) {
	    cout << "Creating crtc " << crtc_id << "\n";
	    crtcs[crtc_id] = new CrtcInfo(display, res, crtc_id);
	} else {
	    crtcs[crtc_id]->update_info(display, res);
	}
    }
}

list<RROutput> RandrInfo::get_outputs()
{
    list<RROutput> ret;

    BOOST_FOREACH(OutputMap::value_type o_it, outputs) {
	ret.push_back(o_it.first);
    }
    return ret;
}

RRMode RandrInfo::get_output_mode(RROutput output_id)
{
    if (outputs[output_id]->crtc == None)
	return None;
    else
	return OUTPUT_MODE(output_id);
}

Rotation RandrInfo::get_output_rotation(RROutput output_id)
{
    if (outputs[output_id]->crtc == None)
	return RR_Rotate_0;
    else
	return crtcs[OUTPUT_CRTC(output_id)]->rotation;
}

list<RRMode> RandrInfo::get_output_modes(RROutput output_id)
{
    list<RRMode> ret;

    BOOST_FOREACH(RRMode mode, outputs[output_id]->modes) {
	ret.push_back(mode);
    }
    return ret;
}

string RandrInfo::get_output_name(RROutput output_id)
{
    string ret = OUTPUT_NAME(output_id);
    return ret;
}

bool RandrInfo::output_is_connected(RROutput output_id)
{
    if (outputs[output_id]->connection == RR_Connected)
	return true;
    else
	return false;
}

string RandrInfo::get_mode_name(RRMode mode_id)
{
    string ret = MODE_NAME(mode_id);
    return ret;
}

list<RRCrtc> RandrInfo::get_crtcs()
{
    list<RRCrtc> ret;

    BOOST_FOREACH(CrtcMap::value_type c_it, crtcs) {
	ret.push_back(c_it.first);
    }
    return ret;
}

RRCrtc RandrInfo::get_output_crtc(RROutput output_id)
{
    return OUTPUT_CRTC(output_id);
}

bool RandrInfo::last_active_crtc(RRCrtc crtc)
{
    int active_crtcs = 0;

    if (crtc == None) {
	cout << "last_active_crtc received a None crtc\n";
	exit(1);
    }

    if (crtcs[crtc]->outputs.size() == 0) {
	// crtc is not active
	return false;
    }

    BOOST_FOREACH(CrtcMap::value_type c_it, crtcs) {
	if (c_it.second->outputs.size() != 0) {
	    active_crtcs++;
	}
    }

    if (active_crtcs == 1)
	return true;
    else
	return false;
}

void RandrInfo::register_for_events(Window win)
{
    event_window = win;
    XRRSelectInput(display, win, RRCrtcChangeNotifyMask |
				 RROutputChangeNotifyMask);

    wanted_x11_event = xrr_event_base + RRNotify;
}

void RandrInfo::hotplug_output(RROutput output, Connection connection)
{
    unsigned int connected_outputs = 0;
    unsigned int used_crtcs = 0;
    unsigned int desired_width, desired_height;

    OutputToModeMap output_to_mode;
    OutputToRotationMap output_to_rotation;

    update_info();

    if (connection == RR_Connected) {
	cout << "Output " << OUTPUT_NAME(output) << " connected\n";

	// Don't activate more outputs than crtcs
	// XXX: this check could go inside set_config (remove from YargWindow
	// too)
	BOOST_FOREACH(CrtcMap::value_type c_it, crtcs) {
	    if (c_it.second->mode != None) {
		used_crtcs++;
	    }
	}
	if (used_crtcs >= crtcs.size()) {
	    cout << "No more available crtcs\n";
	    return;
	}

    } else if (connection == RR_Disconnected) {
	cout << "Output " << OUTPUT_NAME(output) << " disconnected\n";

	// Do nothing if the disconnected monitor was the last one
	BOOST_FOREACH(OutputMap::value_type o_it, outputs) {
	    if (OUTPUT_CRTC(o_it.first) != None) {
		connected_outputs++;
	    }
	}
	if (connected_outputs == 0) {
	    cout << "No more monitors connected. Doing nothing.\n";
	    return;
	}

    } else { // RR_UnknownConnection
	cout << "Output " << OUTPUT_NAME(output) << " connection is unknown\n";
	return;
    }

    // Initialize other outputs
    BOOST_FOREACH(OutputMap::value_type o_it, outputs) {
	if (OUTPUT_CRTC(o_it.first) != None) {
	    output_to_mode[o_it.first] = OUTPUT_MODE(o_it.first);
	    output_to_rotation[o_it.first] = crtcs[OUTPUT_CRTC(o_it.first)]->rotation;
	} else {
	    output_to_mode[o_it.first] = None;
	    output_to_rotation[o_it.first] = RR_Rotate_0;
	}
    }

    output_to_mode[output] = None;

    if (connection == RR_Connected) {
	// Try to clone someone else
	BOOST_FOREACH(OutputMap::value_type o_it, outputs) {

	    if (OUTPUT_CRTC(o_it.first) == None) {
		continue;
	    }

	    desired_width = modes[OUTPUT_MODE(o_it.first)]->width;
	    desired_height = modes[OUTPUT_MODE(o_it.first)]->height;

	    // Check if we support this mode
	    BOOST_FOREACH(RRMode m_it, outputs[output]->modes) {
		if (modes[m_it]->width == desired_width &&
		    modes[m_it]->height == desired_height) {
		    output_to_mode[output] = m_it;
		    break;
		}
	    }

	    if (output_to_mode[output] != None) {
		break;
	    }
	}

	if (output_to_mode[output] == None) {
	    output_to_mode[output] = outputs[output]->preferred_mode;
	}
    }

    set_config(output_to_mode, output_to_rotation,
	       desktop_is_extended(), get_outputs_in_order());
}

void RandrInfo::real_process_event(XEvent *ev)
{
    XRROutputChangeNotifyEvent *out_ev;

    switch (((XRRNotifyEvent*)ev)->subtype) {
#if 0
    case RRNotify_CrtcChange:
	cout << "CrtcChange event\n";
	break;
#endif
    case RRNotify_OutputChange:
	//cout << "OutputChange event\n";

	out_ev = (XRROutputChangeNotifyEvent *)ev;

	if (outputs[out_ev->output]->connection != out_ev->connection) {
	    hotplug_output(out_ev->output, out_ev->connection);
	}
	break;
#if 0
    case RRNotify_OutputProperty:
	//cout << "OutputProperty event\n";
	break;
#endif
    }
}

void RandrInfo::process_event(XEvent *ev)
{
    //  XXX: call XRRUpdateConfiguration (and subscribe for
    //  RRScreenChangeNotify)

    cout << "randr event received!\n";

    real_process_event(ev);

    XEvent other;
    while (XCheckTypedWindowEvent(display, event_window, wanted_x11_event,
				  &other)) {
	cout << "(skipping randr event...)\n";
	real_process_event(&other);
    }

    update_info();
}

RROutput RandrInfo::get_output_from_name(string name)
{
    BOOST_FOREACH(OutputMap::value_type o_it, outputs) {
	if (o_it.second->name == name) {
	    return o_it.first;
	}
    }
    return None;
}

RRMode RandrInfo::get_mode_from_name(string name)
{
    BOOST_FOREACH(ModeMap::value_type m_it, modes) {
	if (m_it.second->name == name) {
	    return m_it.first;
	}
    }
    return None;
}

// I'm using this instead of DisplayWidth() and DisplayHeight() because these
// are macros, and they are only updated after we receive events notifying
// screen size changes and call XRRUpdateConfiguration. Calling
// XGetGeometry is an alternative. Another option would be to call
// QApplication::processEvents() to make qt get the events, process them and
// then update our Display{Width,Height} macros
unsigned int RandrInfo::get_screen_info(ScreenInfo info)
{
    Window root_ret;
    int x_ret, y_ret;
    unsigned int width_ret, height_ret, border_width_ret, depth_ret;

    XGetGeometry(display, RootWindow(display, screen_nbr), &root_ret,
		 &x_ret, &y_ret, &width_ret, &height_ret, &border_width_ret,
		 &depth_ret);

    switch (info) {
	case Width:
	    return width_ret;
	case Height:
	    return height_ret;
	default:
	    cerr << "Invalid argument for get_screen_info\n";
	    exit(1);
    }
}

// XXX: we're using unsigned as arguments to signed
void RandrInfo::set_screen_size(unsigned int width, unsigned int height)
{
    XRRSetScreenSize(display, RootWindow(display, screen_nbr), width, height,
		     DisplayWidthMM(display, screen_nbr),
		     DisplayHeightMM(display, screen_nbr));
    // XXX: add some callback to wait for the events that will update xlib's
    // Display{Width,Height} macros?
    // XXX: we're not changing Display{Width,Height}MM!
}

RandrInfo::SetConfigStatus
RandrInfo::set_config(OutputToModeMap output_to_mode,
		      OutputToRotationMap output_to_rotation,
		      bool extend_desktop,
		      list<RROutput> output_layout)
{
    unsigned int current_width, current_height;
    unsigned int new_width = 0, new_height = 0;
    unsigned int tmp_width, tmp_height;
    unsigned int mode_width, mode_height;
    int min_width, min_height, max_width, max_height;

    RRMode mode;
    RROutput output;
    RRCrtc crtc, assigned_crtc;
    Rotation rotation;
    SetConfigStatus ret;
    int curr_x_pos;
    map<RROutput, unsigned int> output_x_pos;
    CrtcToOutputMap crtc_to_output;

    // XXX: we trust our users and belive that output_to_mode contains every
    // possible output once and that output_layout contains the right layouts

    // Check for the new screen size
    current_width = get_screen_info(Width);
    current_height = get_screen_info(Height);

    BOOST_FOREACH(OutputToModeMap::value_type o2m_it, output_to_mode) {

	output = o2m_it.first;
	mode = o2m_it.second;

	if (mode != None) {

	    rotation = output_to_rotation[output];

	    if (SIDE_ROTATED(rotation)) {
		mode_width = modes[mode]->height;
		mode_height = modes[mode]->width;
	    } else {
		mode_width = modes[mode]->width;
		mode_height = modes[mode]->height;
	    }

	    if (mode_height > new_height) {
		new_height = mode_height;
	    }

	    if (extend_desktop) {
		new_width += mode_width;
	    } else {
		if (mode_width > new_width) {
		    new_width = mode_width;
		}
	    }
	}
    }

    XRRGetScreenSizeRange(display, RootWindow(display, screen_nbr),
			  &min_width, &min_height, &max_width, &max_height);

    if (new_width > max_width || new_height > max_height ||
	new_width < min_width || new_height < min_height) {
	return ScreenSizeError;
    }

    // Set origin position for every one
    if (extend_desktop) {
	curr_x_pos = 0;
	BOOST_FOREACH(RROutput o_it, output_layout) {
	    mode = output_to_mode[o_it];
	    if (mode != None) {
		if (SIDE_ROTATED(output_to_rotation[o_it])) {
		    mode_width = modes[mode]->height;
		} else {
		    mode_width = modes[mode]->width;
		}
		output_x_pos[o_it] = curr_x_pos;
		curr_x_pos += mode_width;
	    }
	}
    }

#ifdef OEM_HACKS
    system("kquitapp plasma > /dev/null 2>&1");
    //system("kill -SIGWITHFIRE plasma");
#endif

    cout << "====== Grabbing Xserver ======\n";
    XGrabServer(display);

    // The screen size must big enough to hold all our operations, so if we need
    // to make it bigger, do it now, but if we need to make it smaller, do it in
    // the end
    tmp_width = current_width;
    tmp_height = current_height;

    if (new_width > current_width)
	tmp_width = new_width;
    if (new_height > current_height)
	tmp_height = new_height;

    if (tmp_width != current_width || tmp_height != current_height) {
	cout << "Changing screen size from " << current_width << "x"
	     << current_height << " to " << tmp_width << "x" << tmp_height
	     << "\n";
	set_screen_size(tmp_width, tmp_height);

	// Now update according to our changes:
	current_width = tmp_width;
	current_height = tmp_height;
    }

    BOOST_FOREACH(CrtcMap::value_type c_it, crtcs) {
	crtc_to_output[c_it.first] = None;
    }

    BOOST_FOREACH(OutputToModeMap::value_type o2m_it, output_to_mode) {
	output = o2m_it.first;
	mode = o2m_it.second;

	if (mode == None) {
	    continue;
	}

	// Try to keep already-assigned things unchanged
	if (OUTPUT_CRTC(output) != None &&
	    crtc_to_output[OUTPUT_CRTC(output)] == None) {
	    assigned_crtc = OUTPUT_CRTC(output);
	    crtc_to_output[assigned_crtc] = output;
	} else {
	    // Try to find a free crtc
	    BOOST_FOREACH(CrtcMap::value_type c_it, crtcs) {
		if (crtc_to_output[c_it.first] == None &&
		    !last_active_crtc(c_it.first)) {
		    crtc_to_output[c_it.first] = output;
		    break;
		}
	    }
	}
    }

    BOOST_FOREACH(CrtcMap::value_type c_it, crtcs) {
	crtc = c_it.first;
	output = crtc_to_output[crtc];
	if (output != None) {
	    rotation = output_to_rotation[output];
	    if (!crtc_supports_rotation(crtc, rotation)) {
		ret = RotationError;
		goto done;
	    }
	}
    }

    // Finally use crtc_to_output and output_to_mode to do the changes
    BOOST_FOREACH(CrtcToOutputMap::value_type c2o_it, crtc_to_output) {

	RROutput *outputs_assigned;
	int noutput;
	Status status;
	int x_pos;

	crtc = c2o_it.first;

	if (crtc == None) {
	    // XXX I guess this shouldn't happen
	    cout << "=========== don't assign crtc None to anything\n";
	    continue;
	}

	output = crtc_to_output[crtc];

	if (output != None) {
	    mode = output_to_mode[output];
	    rotation = output_to_rotation[output];
	    outputs_assigned = &output;
	    noutput = 1;
	} else {
	    mode = None;
	    outputs_assigned = NULL;
	    noutput = 0;
	    rotation = RR_Rotate_0;
	}

	x_pos = output_x_pos[output];

	cout << "Assigning crtc " << crtc << " to output " << output
	     << " and mode " << mode << " at x_pos " << x_pos
	     << " with rotation " << rotation << "\n";

	if (crtcs[crtc]->outputs.size() == 1 &&
	    crtcs[crtc]->outputs.front() == output &&
	    crtcs[crtc]->mode == mode &&
	    crtcs[crtc]->x == x_pos &&
	    crtcs[crtc]->rotation == rotation) {
	    cout << "I don't need to do anything for crtc " << crtc << "\n";
	} else {
	    status = XRRSetCrtcConfig(display, res, crtc, res->timestamp, x_pos, 0,
				      mode, rotation, outputs_assigned, noutput);
	    if (status != RRSetConfigSuccess) {
		cout << "XRRSetCrtcConfig returned " << status << "\n";
		ret = SetCrtcConfigError;
		goto done;
	    }
	    update_info();
	}
    }
    //

    if (new_width != current_width || new_height != current_height) {
	cout << "Changing screen size from " << current_width << "x"
	     << current_height << " to " << new_width << "x" << new_height
	     << "\n";
	set_screen_size(new_width, new_height);
    }

    ret = SetConfigSuccess;

done:
    XSync(display, False);
    XUngrabServer(display);
    cout << "===== Ungrabbing Xserver =====\n";

#ifdef OEM_HACKS
    XSync(display, False); // make sure we ungrab, plasma needs it to quit
    system("while (pidof plasma); do sleep 1; done; setsid plasma > /dev/null 2>&1");
#endif

    return ret;
}

bool RandrInfo::desktop_is_extended()
{
    bool ret = false;

    BOOST_FOREACH(CrtcMap::value_type c_it, crtcs) {
	if (c_it.second->mode != None && c_it.second->x != 0) {
	    ret = true;
	}
    }

    return ret;
}

bool RandrInfo::crtc_supports_rotation(RRCrtc crtc, Rotation rotation)
{
    return crtcs[crtc]->rotation_is_supported(rotation);
}

list<RROutput> RandrInfo::get_outputs_in_order()
{
    list<RROutput> ret;
    int curr_x, next_x, output_x;

    if (!desktop_is_extended()) {
	return ret;
    }

    // If desktop is extended and there are 2 outputs with the same X, this code
    // will put both of them at the list, one after the other. This might cause
    // errors in other parts of yarg. Btw, this shouldn't happen for now

    curr_x = 0;
    while (1) {
	next_x = -1;
	BOOST_FOREACH(OutputMap::value_type o_it, outputs) {
	    if (OUTPUT_CRTC(o_it.first) != None) {
		output_x = crtcs[OUTPUT_CRTC(o_it.first)]->x;
		if (output_x == curr_x) {
		    //cout << "adding " << OUTPUT_NAME(o_it.first) << " to list\n";
		    ret.push_back(o_it.first);
		} else if (next_x > output_x ||
			   (next_x == -1 && output_x > curr_x)) {
		    next_x = output_x;
		}
	    }
	}

	if (next_x == -1) {
	    break;
	}
	curr_x = next_x;
    }

    return ret;
}
