/*
 * Copyright (C) 2010 - Paulo Ricardo Zanoni <pzanoni@mandriva.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 */

#include <iostream>

#include <QtGui>
#include <QX11Info>

#include "YargWindow.h"
#include "OutputWidget.h"
#include "RandrInfo.h"

YargWindow::YargWindow(QWidget *parent)
    : QWidget(parent)
{

    QVBoxLayout *vert_lay = new QVBoxLayout(this);

    QPushButton *apply_btn = new QPushButton(tr("Apply"));
    QPushButton *update_btn = new QPushButton(tr("Update info"));
    QPushButton *hide_btn = new QPushButton(tr("Hide"));

    QMenu *tray_menu = new QMenu("Yarg", this);
    QAction *quit_action = new QAction(tr("Quit"), this);

    x11_info = new QX11Info();

    tray_icon = new QSystemTrayIcon();
    icon = new QIcon("icons:yarg.png");
    config_file = new QFile(QDir::homePath() + "/.config/yarg");

    randr_info = new RandrInfo(x11_info->display(),
                               x11_info->screen());

    randr_info->register_for_events(winId());

    foreach(RROutput output, randr_info->get_outputs()) {
	output_widgets.append(new OutputWidget(randr_info, output));
	vert_lay->addWidget(output_widgets.last());

	connect(output_widgets.last(),
		SIGNAL(output_changed(RROutput, RRMode, Rotation)),
		this, SLOT(ow_output_changed(RROutput, RRMode, Rotation)));
    }

    extend_widget = new ExtendDesktopWidget();

    vert_lay->addWidget(extend_widget);
    vert_lay->addWidget(apply_btn);
    vert_lay->addWidget(update_btn);
    vert_lay->addWidget(hide_btn);

    connect(apply_btn, SIGNAL(clicked()), this, SLOT(apply_btn_click()));
    connect(update_btn, SIGNAL(clicked()), this, SLOT(update_btn_click()));
    connect(hide_btn, SIGNAL(clicked()), this, SLOT(hide_btn_click()));
    connect(quit_action, SIGNAL(triggered()), qApp, SLOT(quit()));
    connect(tray_icon, SIGNAL(activated(QSystemTrayIcon::ActivationReason)),
	    this, SLOT(tray_icon_activated(QSystemTrayIcon::ActivationReason)));

    load_config();

    setWindowTitle(tr("Yet Another Randr Gui"));
    setWindowIcon(*icon);

    setFixedSize(vert_lay->minimumSize());

    tray_menu->addAction(quit_action);
    tray_icon->setIcon(*icon);
    tray_icon->setContextMenu(tray_menu);
    tray_icon->show();
}

YargWindow::~YargWindow()
{

    foreach(OutputWidget *ow, output_widgets) {
	delete ow;
    }
    delete randr_info;
}

void YargWindow::hide_btn_click()
{
    hide();
}

void YargWindow::update_btn_click()
{
    randr_info->update_info();
    update_info();
}

void YargWindow::update_info(bool do_save_config)
{
    // possible cases for this function:
    //  - update button clicked (randr was told to update)
    //  - window was told to show() via systray (randr was told to update)
    //  - window was told to show() via SIGUSR1 (we should tell randr to update,
    //    but we don't)
    //  - event received (randr updated itself)
    //  - we changed resolution (randr updated itself)
    std::cout << "-- updating YargWindow info --\n";
    //randr_info->update_info();

    foreach(OutputWidget *ow, output_widgets) {
	ow->update_info();
    }

    update_extend_widget();

    if (do_save_config) {
	save_config();
    }
}

void YargWindow::apply_btn_click()
{
    apply_config(true);
}

void YargWindow::apply_config(bool do_save_config)
{
    bool extend_desktop = extend_widget->is_extended();

    configure_outputs(extend_desktop);
    update_info(do_save_config);
}

void YargWindow::configure_outputs(bool extend_desktop)
{
    unsigned int enabled_outputs = 0;
    std::list<RRCrtc> crtcs;
    unsigned int ncrtc;
    RandrInfo::OutputToModeMap output_to_mode;
    RandrInfo::OutputToRotationMap output_to_rotation;
    RandrInfo::SetConfigStatus status;
    RROutput output;

    foreach(OutputWidget *ow, output_widgets) {
	if (ow->get_selected_mode() != None) {
	    enabled_outputs++;
	}
    }

    crtcs = randr_info->get_crtcs();
    ncrtc = crtcs.size();

    std::cout << "enabled outputs: " << enabled_outputs <<
	         " number of crtcs: " << ncrtc << "\n";

    // XXX: don't let more activated outputs than crtcs
    //      this is not necessary, but requires careful implementation
    //      adding this feature could trigger a new minor version =)
    if (enabled_outputs > ncrtc) {
	QString text;
	//QString informative_text;
	//text.append(tr("There are more activated outputs than CRT controllers."));
	text.append(tr("You may only use up to %1 video output(s).").arg(ncrtc));
	error_dialog(text);
	return;

    } else if (enabled_outputs == 0) {
	QString text;
	error_dialog(tr("You must let at least one active video output."));
	return;
    }

    foreach(OutputWidget *ow, output_widgets) {
	output = ow->get_output();
	output_to_mode[output] = ow->get_selected_mode();
	output_to_rotation[output] = ow->get_selected_rotation();
    }

#ifdef OEM_HACKS
    tray_icon->hide();
#endif

    QList<RROutput> list;
    list = extend_widget->get_outputs_in_order();
    status = randr_info->set_config(output_to_mode,
				    output_to_rotation,
				    extend_desktop,
				    list.toStdList());

#ifdef OEM_HACKS
    while (!QSystemTrayIcon::isSystemTrayAvailable()) {
	std::cout << "Waiting for the system tray...\n";
	sleep(1);
    }
    tray_icon->setIcon(*icon);
    tray_icon->show();
#endif

    if (status == RandrInfo::ScreenSizeError) {
	error_dialog(tr("Screen size is too big"));
    } else if (status == RandrInfo::SetCrtcConfigError) {
	error_dialog(tr("Error setting configuration"),
		     tr("Maybe another application is trying to change the "
			"screen configuration. Please try again."));
    } else if (status == RandrInfo::RotationError) {
	error_dialog(tr("Selected rotation is not supported by display"));
    }

}

bool YargWindow::x11Event(XEvent *event)
{
    if (event->type == randr_info->wanted_x11_event) {
	randr_info->process_event(event);
	update_info();
	return true;
    }
    return false; // qt will process it
}

void YargWindow::closeEvent(QCloseEvent *event)
{
    hide();

    // FIXME: if yarg is not minimized to system tray and the user tries to
    // logout using KDE (untested in other environments), we receive a
    // closeEvent, and if we ignore() it, the user will get a message telling
    // him that yarg is preventing the logout. The difference between this
    // "logout" event and the "close button click" event is that the first is
    // not spontaneous.  Maybe there is another way to check if the close event
    // is due to a "logout", but for now this code is working.
    if (event->spontaneous()) {
	event->ignore();
    } else {
	event->accept();
    }

}

void YargWindow::tray_icon_activated(QSystemTrayIcon::ActivationReason reason)
{
    if (reason == QSystemTrayIcon::Context) {
	tray_icon->contextMenu()->exec();
    } else {
	if (isHidden()) {
	    randr_info->update_info();
	    update_info();
	    show();
	} else {
	    hide();
	}
    }
}

void YargWindow::error_dialog(QString text, QString informative_text)
{
    QMessageBox dialog;
    if (text == NULL) {
	dialog.setText(tr("Failed!"));
    } else {
	dialog.setText(text);
    }
    if (informative_text != NULL) {
	dialog.setInformativeText(informative_text);
    }
    dialog.setIcon(QMessageBox::Warning);
    dialog.setWindowTitle(tr("Error"));
    dialog.setWindowIcon(*icon);
    dialog.exec();
}


void YargWindow::save_config()
{
    RROutput output;
    RRMode mode;
    Rotation rotation;
    std::string output_name;
    std::string mode_name;

    std::cout << "saving config\n";

    if (!config_file->open(QIODevice::WriteOnly | QIODevice::Text)) {
	std::cout << "Can't open config file "
		  << config_file->fileName().toStdString() << "\n";
    }

    QTextStream out(config_file);

    foreach(OutputWidget *ow, output_widgets) {
	output = ow->get_output();
	mode = ow->get_selected_mode();

	output_name = randr_info->get_output_name(output);

	if (mode != None) {
	    mode_name = randr_info->get_mode_name(mode);
	    rotation = randr_info->get_output_rotation(output);
	} else {
	    mode_name = "None";
	    rotation = RR_Rotate_0;
	}

	out << output_name.c_str() << "\n"
	    << mode_name.c_str() << "\n"
	    << rotation << "\n";
    }

    if (extend_widget->is_extended()) {
	out << "extended\n";

	QList<RROutput> outputs_in_order;
	outputs_in_order = extend_widget->get_outputs_in_order();
	for(int i = 0; i < outputs_in_order.size(); i++) {
	    output_name = randr_info->get_output_name(outputs_in_order[i]);
	    out << output_name.c_str() << "\n";
	}
    }

    config_file->close();
}

void YargWindow::load_config()
{
    QString output_name;
    QString mode_name;
    RROutput output_id;
    RRMode mode_id;
    Rotation rotation;

    bool reading_extended_info = false;
    QList<QPair<RROutput, QString*>*> outputs_in_order;
    QPair<RROutput, QString*> *pair;

    std::cout << "loading config\n";

    if (!config_file->open(QIODevice::ReadOnly | QIODevice::Text)) {
	std::cout << "Can't open config file "
		  << config_file->fileName().toStdString() << "\n";
	return;
    }

    QTextStream in(config_file);

    while (!in.atEnd()) {
	output_name = in.readLine();

	if (output_name == "extended") {
	    extend_widget->set_extended(true);
	    reading_extended_info = true;
	    continue;
	}

	output_id = randr_info->get_output_from_name(output_name.toStdString());

	if (!reading_extended_info) {

	    mode_name = in.readLine();
	    mode_id = randr_info->get_mode_from_name(mode_name.toStdString());
	    rotation = in.readLine().trimmed().toUInt();


	    std::cout << "name: " << output_name.toStdString() << "(" << output_id
		      << ")" << " mode: " << mode_name.toStdString() << "("
		      << mode_id << ")\n";

	    foreach(OutputWidget *ow, output_widgets) {
		if (ow->get_output() == output_id) {
		    ow->set_selected_mode(mode_id);
		    ow->set_selected_rotation(rotation);
		}
	    }
	} else {
	    pair = new QPair<RROutput, QString*>(output_id, &output_name);
	    outputs_in_order << pair;
	}

    }

    if (reading_extended_info) {
	extend_widget->set_items(outputs_in_order);

	while (!outputs_in_order.isEmpty()) {
	    delete outputs_in_order.takeFirst();
	}
    }

    apply_config(false);

    config_file->close();
}

void YargWindow::ow_output_changed(RROutput output, RRMode mode,
				   Rotation rotation)
{
    std::cout << "----- ow_output_changed\n";

    if (mode == None) {
	extend_widget->remove_output(output);
    } else {
	std::string output_name = randr_info->get_output_name(output);
	extend_widget->add_output(output, output_name);
    }
}

void YargWindow::update_extend_widget()
{
    QList<QPair<RROutput, QString*>* > used_output_names;
    QPair<RROutput, QString*> *pair;
    QString *string;
    std::string output_name;

    std::cout << "----- update_extend_widget\n";

    if (randr_info->desktop_is_extended()) {
	extend_widget->set_extended(true);

	foreach(RROutput output, randr_info->get_outputs_in_order()) {
	    output_name = randr_info->get_output_name(output);
	    string = new QString(output_name.c_str());
	    pair = new QPair<RROutput, QString*>(output, string);
	    used_output_names << pair;
	}

	extend_widget->set_items(used_output_names);

	for(int i = 0; i < used_output_names.size(); i++) {
	    delete used_output_names.at(i)->second;
	    delete used_output_names.at(i);
	}

    } else {
	extend_widget->set_extended(false);
    }
}
